# pylint: disable=protected-access
import logging
import traceback
from datetime import datetime
from decimal import Decimal
from typing import Optional, Sequence, Any, Dict
from django.contrib import messages
from django.contrib.admin import SimpleListFilter
from django.contrib.messages import add_message, INFO
from django.db import models
from django.db.models.functions import Coalesce
from django import forms
from django.shortcuts import render
from django.urls import reverse, ResolverMatch, path
from django.utils.formats import date_format
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.contrib.admin import widgets
from django.utils.text import format_lazy, capfirst
from django.utils.timezone import now
from jacc.format import align_lines
from jacc.forms import ReverseChargeForm
from jacc.models import (
    Account,
    AccountEntry,
    Invoice,
    AccountType,
    EntryType,
    AccountEntrySourceFile,
    INVOICE_STATE,
    AccountEntryNote,
    INVOICE_NOT_DUE_YET,
    INVOICE_DUE,
    INVOICE_LATE,
)
from django.conf import settings
from django.contrib import admin
from django.core.exceptions import ValidationError
from django.db.models import QuerySet, Sum, Count
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _
from jacc.settle import settle_assigned_invoice
from jutil.admin import ModelAdminBase, admin_log
from jutil.format import choices_label, dec2
from jutil.model import clone_model

logger = logging.getLogger(__name__)


def refresh_cached_fields(modeladmin, request, qs):  # pylint: disable=unused-argument
    n_count = 0
    for obj in qs.order_by("id").distinct():
        try:
            obj.update_cached_fields()
            n_count += 1
        except Exception as exc:
            add_message(request, messages.ERROR, f"{obj}: {exc}")
    add_message(request, messages.SUCCESS, _("Cached fields refreshed ({})").format(n_count))


def summarize_account_entries(modeladmin, request, qs):  # pylint: disable=unused-argument
    # {total_count} entries:
    # {amount1} {currency} x {count1} = {total1} {currency}
    # {amount2} {currency} x {count2} = {total2} {currency}
    # Total {total_amount} {currency}
    e_type_entries = list(qs.distinct("type").order_by("type"))
    total_debits = Decimal("0.00")
    total_credits = Decimal("0.00")
    lines = [
        "<pre>",
        _("({total_count} account entries)").format(total_count=qs.count()),
        "",
        "|".join([str(_("entry type")), str(_("count")), str(_("credit")), str(_("debit"))]),
    ]

    for e_type_entry in e_type_entries:
        assert isinstance(e_type_entry, AccountEntry)
        e_type = e_type_entry.type
        assert isinstance(e_type, EntryType)

        qs2 = qs.filter(type=e_type)
        res_debit = qs2.filter(amount__gt=0).aggregate(total=Coalesce(Sum("amount"), Decimal("0.00")), n=Count("amount"))
        res_credit = qs2.filter(amount__lt=0).aggregate(total=Coalesce(Sum("amount"), Decimal("0.00")), n=Count("amount"))
        lines.append(
            "{name}|{n}|{cr:.2f}|{dr:.2f}".format(name=e_type.name, n=res_credit["n"] + res_debit["n"], cr=-res_credit["total"], dr=res_debit["total"])
        )
        total_debits += res_debit["total"]
        total_credits += res_credit["total"]

    lines.append("")
    lines.append(
        _("Total debits {total_debits:.2f} | - total credits {total_credits:.2f} | = {total_amount:.2f}").format(
            total_debits=total_debits, total_credits=-total_credits, total_amount=total_debits + total_credits
        )
    )
    lines = align_lines(lines, "|")
    messages.add_message(request, INFO, format_html("<br>".join(lines)), extra_tags="safe")


class SettlementAccountEntryFilter(SimpleListFilter):
    title = _("settlement")
    parameter_name = "settlement"

    def lookups(self, request, model_admin):
        return [
            ("S", _("settlement")),
            ("P", _("payment")),
            ("N", _("non-payment settlement")),
            ("0", _("not settlement")),
        ]

    def queryset(self, request, queryset):
        val = self.value()
        if val:
            if val == "S":
                queryset = queryset.filter(parent=None, type__is_settlement=True)
            if val == "P":
                queryset = queryset.filter(parent=None, type__is_settlement=True, type__is_payment=True)
            if val == "N":
                queryset = queryset.filter(parent=None, type__is_settlement=True, type__is_payment=False)
            elif val == "0":
                queryset = queryset.exclude(type__is_settlement=True)
        return queryset


class EntryTypeAccountEntryFilter(SimpleListFilter):
    title = _("account entry type")
    parameter_name = "type"

    def lookups(self, request, model_admin):
        a = []
        for e in EntryType.objects.all().filter(is_settlement=True).order_by("name"):
            a.append((e.code, e.name))
        return a

    def queryset(self, request, queryset):
        val = self.value()
        if val:
            queryset = queryset.filter(type__code=val)
        return queryset


class AccountTypeAccountEntryFilter(SimpleListFilter):
    title = _("account type")
    parameter_name = "atype"

    def lookups(self, request, model_admin):
        a = []
        for e in AccountType.objects.all().order_by("name"):
            a.append((e.code, e.name))
        return a

    def queryset(self, request, queryset):
        val = self.value()
        if val:
            queryset = queryset.filter(account__type__code=val)
        return queryset


def add_reverse_charge(modeladmin, request, qs):
    assert hasattr(modeladmin, "reverse_charge_form")
    assert hasattr(modeladmin, "reverse_charge_template")

    user = request.user
    cx: Dict[str, Any] = {}  # type: ignore
    try:
        if qs.count() != 1:
            raise ValidationError(_("Exactly one account entry must be selected"))
        e = qs.first()
        assert isinstance(e, AccountEntry)
        if e.amount is None or dec2(e.amount) == Decimal("0.00"):
            raise ValidationError(_("Exactly one account entry must be selected"))

        form_cls = modeladmin.reverse_charge_form  # Type: ignore
        form = form_cls()
        default_description = form.get_initial_for_field(form.fields["description"], "description") if hasattr(form, "get_initial_for_field") else ""
        initial = {
            "amount": -e.amount,
            "description": default_description,
        }
        if e.description:
            initial["description"] += " / {}".format(e.description)
        cx = {
            "qs": qs,
            "original": e,
        }

        if "save" in request.POST:
            cx["form"] = form = form_cls(request.POST, initial=initial)
            if not form.is_valid():
                raise ValidationError(form.errors)
            timestamp = form.cleaned_data["timestamp"]
            amount = form.cleaned_data["amount"]
            description = form.cleaned_data["description"]
            reverse_e = clone_model(
                e,
                parent=e.parent,
                amount=amount,
                description=description,
                timestamp=timestamp,
                commit=False,
                created=now(),
            )
            reverse_e.full_clean()
            reverse_e.save()
            messages.info(request, "{} {}".format(reverse_e, _("created")))
            admin_log([e, reverse_e], f"Added {reverse_e} to reverse {e}: {description}", who=user)
        else:
            cx["form"] = form = form_cls(initial=initial)
            return render(request, modeladmin.reverse_charge_template, context=cx)
    except ValidationError as e:
        if cx:
            return render(request, modeladmin.reverse_charge_template, context=cx)
        messages.error(request, "{}\n".join(e.messages))
    except Exception as e:
        logger.error("add_reverse_charge: %s", traceback.format_exc())
        messages.error(request, "{}".format(e))
    return None


class AccountEntryAdminForm(forms.ModelForm):
    def clean(self):
        if self.instance.archived:
            raise ValidationError(_("cannot.modify.archived.account.entry"))
        return super().clean()


class AccountEntryNoteInline(admin.TabularInline):
    model = AccountEntryNote
    fk_name = "account_entry"
    verbose_name = _("account entry note")
    verbose_name_plural = _("account entry notes")
    extra = 1
    can_delete = True
    fields = [
        "created",
        "note",
        "created_by",
    ]
    readonly_fields = [
        "created",
        "created_by",
    ]
    formfield_overrides = {
        models.TextField: {"widget": widgets.AdminTextareaWidget(attrs={"rows": 3})},
    }


class AccountEntryAdmin(ModelAdminBase):
    form = AccountEntryAdminForm
    date_hierarchy = "timestamp"
    list_per_page = 50
    reverse_charge_form = ReverseChargeForm
    reverse_charge_template = "admin/jacc/accountentry/reverse_entry.html"
    actions = [
        summarize_account_entries,
        add_reverse_charge,
    ]
    list_display = [
        "id",
        "timestamp",
        "account_link",
        "type",
        "amount",
        "parent",
    ]
    raw_id_fields = [
        "account",
        "source_file",
        "type",
        "parent",
        "source_invoice",
        "settled_invoice",
        "settled_item",
        "parent",
    ]
    ordering = [
        "-id",
    ]
    search_fields = [
        "=id",
        "=amount",
        "description",
    ]
    fields = [
        "id",
        "account",
        "timestamp",
        "created",
        "last_modified",
        "type",
        "description",
        "amount",
        "source_file",
        "source_invoice",
        "settled_invoice",
        "settled_item",
        "parent",
        "archived",
    ]
    readonly_fields = [
        "id",
        "created",
        "last_modified",
        "balance",
        "source_invoice_link",
        "settled_invoice_link",
        "settled_item_link",
        "archived",
    ]
    list_filter = [
        SettlementAccountEntryFilter,
        EntryTypeAccountEntryFilter,
        AccountTypeAccountEntryFilter,
        "archived",
    ]
    inlines = [
        AccountEntryNoteInline,
    ]
    account_admin_change_view_name = "admin:jacc_account_change"
    invoice_admin_change_view_name = "admin:jacc_invoice_change"
    accountentrysourcefile_admin_change_view_name = "admin:jacc_accountentrysourcefile_change"
    accountentry_admin_change_view_name = "admin:jacc_accountentry_change"
    allow_add = False
    allow_delete = False
    allow_change = False

    def fill_extra_context(self, request: HttpRequest, extra_context: Optional[Dict[str, Any]]):
        extra_context = extra_context or {}
        rm = request.resolver_match
        assert isinstance(rm, ResolverMatch)
        pk = rm.kwargs.get("account_id") or request.GET.get("account")
        if pk:
            extra_context["account"] = Account.objects.filter(id=pk).first()
        return extra_context

    def add_view(self, request: HttpRequest, form_url="", extra_context=None):
        return super().add_view(request, form_url, self.fill_extra_context(request, extra_context))

    def change_view(self, request: HttpRequest, object_id, form_url="", extra_context=None):
        return super().change_view(request, object_id, form_url, self.fill_extra_context(request, extra_context))

    def custom_changelist_view(self, request: HttpRequest, extra_context=None, **kwargs):  # pylint: disable=unused-argument
        return self.changelist_view(request, self.fill_extra_context(request, extra_context))

    def source_file_link(self, obj):
        assert isinstance(obj, AccountEntry)
        if not obj.source_file:
            return ""
        admin_url = reverse(self.accountentrysourcefile_admin_change_view_name, args=(obj.source_file.id,))
        return format_html("<a href='{}'>{}</a>", mark_safe(admin_url), obj.source_file)

    source_file_link.admin_order_field = "source_file"  # type: ignore
    source_file_link.short_description = _("account entry source file")  # type: ignore

    def account_link(self, obj):
        admin_url = reverse(self.account_admin_change_view_name, args=(obj.account.id,))
        return format_html("<a href='{}'>{}</a>", mark_safe(admin_url), obj.account)

    account_link.admin_order_field = "account"  # type: ignore
    account_link.short_description = _("account")  # type: ignore

    def source_invoice_link(self, obj):
        if not obj.source_invoice:
            return ""
        admin_url = reverse(self.invoice_admin_change_view_name, args=(obj.source_invoice.id,))
        return format_html("<a href='{}'>{}</a>", mark_safe(admin_url), obj.source_invoice)

    source_invoice_link.admin_order_field = "source_invoice"  # type: ignore
    source_invoice_link.short_description = _("source invoice")  # type: ignore

    def settled_invoice_link(self, obj):
        if not obj.settled_invoice:
            return ""
        admin_url = reverse(self.invoice_admin_change_view_name, args=(obj.settled_invoice.id,))
        return format_html("<a href='{}'>{}</a>", mark_safe(admin_url), obj.settled_invoice)

    settled_invoice_link.admin_order_field = "settled_invoice"  # type: ignore
    settled_invoice_link.short_description = _("settled invoice")  # type: ignore

    def settled_item_link(self, obj):
        if not obj.settled_item:
            return ""
        admin_url = reverse(self.accountentry_admin_change_view_name, args=(obj.settled_item.id,))
        return format_html("<a href='{}'>{}</a>", mark_safe(admin_url), obj.settled_item)

    settled_item_link.admin_order_field = "settled_item"  # type: ignore
    settled_item_link.short_description = _("settled item")  # type: ignore

    def parent_link(self, obj):
        if obj.parent is None:
            return ""
        admin_url = reverse(self.accountentry_admin_change_view_name, args=(obj.parent.id,))
        return format_html("<a href='{}'>{}</a>", mark_safe(admin_url), obj.parent)

    parent_link.admin_order_field = "parent"  # type: ignore
    parent_link.short_description = _("account.entry.parent")  # type: ignore

    def get_urls(self):
        info = self.model._meta.app_label, self.model._meta.model_name  # noqa
        return [
            path(
                "by-account/<int:account_id>/",
                self.admin_site.admin_view(self.custom_changelist_view),
                name="%s_%s_account_changelist" % info,
            ),
            path(
                "by-source-invoice/<int:source_invoice_id>/",
                self.admin_site.admin_view(self.custom_changelist_view),
                name="%s_%s_source_invoice_changelist" % info,
            ),
            path(
                "by-settled-invoice/<int:settled_invoice_id>/",
                self.admin_site.admin_view(self.custom_changelist_view),
                name="%s_%s_settled_invoice_changelist" % info,
            ),
            path(
                "by-source-file/<int:source_file_id>/",
                self.admin_site.admin_view(self.custom_changelist_view),
                name="%s_%s_sourcefile_changelist" % info,
            ),
        ] + super().get_urls()

    def get_queryset(self, request: HttpRequest):
        qs = super().get_queryset(request)
        rm = request.resolver_match
        assert isinstance(rm, ResolverMatch)
        info = self.model._meta.app_label, self.model._meta.model_name  # noqa
        account_id = rm.kwargs.get("account_id")
        if account_id:
            return qs.filter(account=account_id)
        source_file_id = rm.kwargs.get("source_file_id")
        if source_file_id:
            return qs.filter(source_file=source_file_id)
        source_invoice_id = rm.kwargs.get("source_invoice_id")
        if source_invoice_id:
            return qs.filter(source_invoice=source_invoice_id)
        settled_invoice_id = rm.kwargs.get("settled_invoice_id")
        if settled_invoice_id:
            return qs.filter(settled_invoice=settled_invoice_id)
        return qs

    def save_formset(self, request, form, formset, change):
        if formset.model == AccountEntryNote:
            AccountEntryNoteAdmin.save_account_entry_note_formset(request, form, formset, change)
        else:
            formset.save()


class AccountAdmin(ModelAdminBase):
    list_display = [
        "id",
        "type",
        "name",
        "balance",
        "currency",
        "is_asset",
    ]
    fields: Sequence[str] = ["id", "type", "name", "balance", "currency", "notes"]
    search_fields = [
        "name",
    ]
    readonly_fields = [
        "id",
        "balance",
        "is_asset",
    ]
    raw_id_fields = [
        "type",
    ]
    ordering = [
        "-id",
    ]
    list_filter = [
        "type",
        "type__is_asset",
    ]
    allow_add = True
    allow_delete = True
    list_per_page = 20


class AccountEntryInlineFormSet(forms.BaseInlineFormSet):
    def clean_entries(self, source_invoice: Optional[Invoice], settled_invoice: Optional[Invoice], account: Optional[Account], **kw):
        """This needs to be called from a derived class clean().

        Args:
            source_invoice
            settled_invoice
            account

        Returns:
            None
        """
        for form in self.forms:
            obj = form.instance
            assert isinstance(obj, AccountEntry)
            if account is not None:
                obj.account = account
            obj.source_invoice = source_invoice
            obj.settled_invoice = settled_invoice
            if obj.parent is not None:
                if obj.amount is None:
                    obj.amount = obj.parent.amount
                if obj.type is None:
                    obj.type = obj.parent.type
                if obj.amount is not None and obj.parent.amount is not None:
                    if obj.amount > obj.parent.amount > Decimal(0) or obj.amount < obj.parent.amount < Decimal(0):
                        raise ValidationError(_("Derived account entry amount cannot be larger than original"))
            elif obj.type is not None:
                if obj.type.is_payment:
                    raise ValidationError(_("Payment settlements must have originating account entry which is also payment"))
            if obj.type is not None and obj.parent is not None and obj.parent.type is not None:
                if obj.type.is_payment != obj.parent.type.is_payment:
                    raise ValidationError(_("Payment settlements must have originating account entry which is also payment"))
            for k, v in kw.items():
                setattr(obj, k, v)


class SingleReceivablesAccountInvoiceItemInlineFormSet(AccountEntryInlineFormSet):
    def clean(self):
        instance = self.instance
        assert isinstance(instance, Invoice)
        receivables_account = Account.objects.get(type__code=settings.ACCOUNT_RECEIVABLES)
        self.clean_entries(instance, None, receivables_account)


class SingleSettlementsAccountSettlementInlineFormSet(AccountEntryInlineFormSet):
    def clean(self):
        instance = self.instance
        assert isinstance(instance, Invoice)
        settlement_account = Account.objects.get(type__code=settings.ACCOUNT_SETTLEMENTS)
        self.clean_entries(None, instance, settlement_account)

    def save(self, commit=True):
        instance = self.instance
        assert isinstance(instance, Invoice)
        entries = super().save(commit)
        settlement_account = Account.objects.get(type__code=settings.ACCOUNT_SETTLEMENTS)
        assert isinstance(settlement_account, Account)
        for e in entries:
            if settlement_account.needs_settling(e):
                settle_assigned_invoice(instance.receivables_account, e, AccountEntry)
        return entries


class InvoiceItemInline(admin.TabularInline):  # TODO: override in app
    model = AccountEntry
    formset = SingleReceivablesAccountInvoiceItemInlineFormSet  # type: ignore  # TODO: override in app
    fk_name = "source_invoice"
    verbose_name = _("invoice items")
    verbose_name_plural = _("invoices items")
    extra = 0
    can_delete = True
    account_entry_change_view_name = "admin:jacc_accountentry_change"
    fields = [
        "id_link",
        "timestamp",
        "type",
        "description",
        "amount",
    ]
    raw_id_fields = [
        "account",
        "type",
        "source_invoice",
        "settled_invoice",
        "settled_item",
        "source_file",
        "parent",
    ]
    readonly_fields = [
        "id_link",
    ]

    def id_link(self, obj):
        if obj and obj.id:
            assert isinstance(obj, AccountEntry)
            admin_url = reverse(self.account_entry_change_view_name, args=(obj.id,))
            return format_html("<a href='{}'>{}</a>", mark_safe(admin_url), obj.id)
        return ""

    id_link.admin_order_field = "id"  # type: ignore
    id_link.short_description = _("id")  # type: ignore

    def get_queryset(self, request):
        queryset = self.model._default_manager.get_queryset().filter(type__is_settlement=False)
        if not self.has_change_permission(request):
            queryset = queryset.none()
        return queryset

    def get_field_queryset(self, db, db_field, request):
        related_admin = self.admin_site._registry.get(db_field.remote_field.model)
        if related_admin and db_field.name == "type":
            return related_admin.get_queryset(request).filter(is_settlement=False).order_by("name")
        return super().get_field_queryset(db, db_field, request)


class InvoiceSettlementInline(admin.TabularInline):  # TODO: override in app
    model = AccountEntry
    formset = SingleSettlementsAccountSettlementInlineFormSet  # type: ignore  # TODO: override in app
    fk_name = "settled_invoice"
    verbose_name = _("settlements")
    verbose_name_plural = _("settlements")
    show_non_settlements = False
    extra = 0
    can_delete = True
    account_entry_change_view_name = "admin:jacc_accountentry_change"  # TODO: override in app
    account_change_view_name = "admin:jacc_account_change"  # TODO: override in app
    fields = [
        "id_link",
        "account_link",
        "timestamp",
        "type",
        "description",
        "amount",
        "parent",
        "settled_item",
    ]
    raw_id_fields = [
        "account",
        "type",
        "source_invoice",
        "settled_invoice",
        "source_file",
        "parent",
        "settled_item",
    ]
    readonly_fields = [
        "id_link",
        "account_link",
        "settled_item",
    ]

    def get_queryset(self, request):
        queryset = self.model._default_manager.get_queryset()
        if not self.show_non_settlements:
            queryset = queryset.filter(type__is_settlement=True)
        if not self.has_change_permission(request):
            queryset = queryset.none()
        return queryset

    def get_field_queryset(self, db, db_field, request):
        related_admin = self.admin_site._registry.get(db_field.remote_field.model)
        if related_admin and db_field.name == "type" and not self.show_non_settlements:
            return related_admin.get_queryset(request).filter(is_settlement=True).order_by("name")
        return super().get_field_queryset(db, db_field, request)

    def id_link(self, obj):
        if obj and obj.id:
            assert isinstance(obj, AccountEntry)
            admin_url = reverse(self.account_entry_change_view_name, args=(obj.id,))
            return format_html("<a href='{}'>{}</a>", mark_safe(admin_url), obj.id)
        return ""

    id_link.admin_order_field = "id"  # type: ignore
    id_link.short_description = _("id")  # type: ignore

    def account_link(self, obj):
        if obj and obj.id:
            assert isinstance(obj, AccountEntry)
            admin_url = reverse(self.account_change_view_name, args=(obj.account.id,))
            return format_html("<a href='{}'>{}</a>", mark_safe(admin_url), obj.account)
        return ""

    account_link.admin_order_field = "account"  # type: ignore
    account_link.short_description = _("account")  # type: ignore


def resend_invoices(modeladmin, request: HttpRequest, queryset: QuerySet):  # pylint: disable=unused-argument
    """Marks invoices with as un-sent.

    Args:
        modeladmin
        request
        queryset

    Returns:

    """
    user = request.user
    for obj in queryset:
        assert isinstance(obj, Invoice)
        admin_log([obj, user], "Invoice id={invoice} marked for re-sending".format(invoice=obj.id), who=user)
    queryset.update(sent=None)


class InvoiceLateDaysFilter(SimpleListFilter):
    title = _("late days")
    parameter_name = "late_days_range"

    def lookups(self, request, model_admin):
        if hasattr(settings, "INVOICE_LATE_DAYS_LIST_FILTER"):
            return settings.INVOICE_LATE_DAYS_LIST_FILTER
        return [
            ("<0", _("late.days.filter.not.due")),
            ("0<7", format_lazy(_("late.days.filter.late.range"), 1, 7)),
            ("7<14", format_lazy(_("late.days.filter.late.range"), 7, 14)),
            ("14<21", format_lazy(_("late.days.filter.late.range"), 14, 21)),
            ("21<28", format_lazy(_("late.days.filter.late.range"), 21, 28)),
            ("28<", format_lazy(_("late.days.filter.late.over.days"), 28)),
        ]

    def queryset(self, request, queryset):
        val = self.value()
        if val:
            begin, end = str(val).split("<")
            if begin:
                queryset = queryset.filter(late_days__gte=int(begin))
            if end:
                queryset = queryset.filter(late_days__lt=int(end))
        return queryset


def summarize_invoice_statistics(modeladmin, request: HttpRequest, qs: QuerySet):  # pylint: disable=unused-argument
    invoice_states = list(state for (state, name) in INVOICE_STATE)

    invoiced_total_amount = Decimal("0.00")
    invoiced_total_count = 0

    lines = [
        "<pre>",
        _("({total_count} invoices)").format(total_count=qs.count()),
    ]
    for state in invoice_states:
        state_name = choices_label(INVOICE_STATE, state)
        qs2 = qs.filter(state=state)

        invoiced = qs2.filter(state=state).aggregate(amount=Coalesce(Sum("amount"), Decimal("0.00")), count=Count("*"))
        invoiced_amount = Decimal(invoiced["amount"])
        invoiced_count = int(invoiced["count"])
        invoiced_total_amount += invoiced_amount
        invoiced_total_count += invoiced_count

        lines.append("{state_name} | x{count} | {amount:.2f}".format(state_name=state_name, amount=invoiced_amount, count=invoiced_count))

    lines.append(_("Total") + " {label} | x{count} | {amount:.2f}".format(label=_("amount"), amount=invoiced_total_amount, count=invoiced_total_count))
    lines.append("</pre>")

    lines = align_lines(lines, "|")
    messages.add_message(request, INFO, format_html("<br>".join(lines)), extra_tags="safe")


class InvoiceStateFilter(SimpleListFilter):
    title = _("state")
    parameter_name = "invoice-state"

    def lookups(self, request, model_admin):
        late_limit_days = settings.LATE_LIMIT_DAYS
        day_abbr = _("day.abbr")
        return [
            ("O", capfirst(_("outstanding.invoice"))),
            (INVOICE_NOT_DUE_YET, mark_safe("&nbsp;" * 2 + _("Not due yet"))),
            ("DL", mark_safe("&nbsp;" * 2 + _("Due") + "&nbsp;/&nbsp;" + _("late"))),
            (INVOICE_DUE, mark_safe("&nbsp;" * 4 + _("Due") + f"&nbsp;({late_limit_days}{day_abbr})")),
            (INVOICE_LATE, mark_safe("&nbsp;" * 4 + _("Late"))),
            ("C", capfirst(_("closed.invoice"))),
        ]

    def queryset(self, request, queryset):
        val = self.value()
        if val:
            if val == "O":
                queryset = queryset.filter(close_date=None)
            elif val == "C":
                queryset = queryset.exclude(close_date=None)
            elif val == "DL":
                queryset = queryset.filter(state__in=[INVOICE_DUE, INVOICE_LATE])
            else:
                queryset = queryset.filter(state=val)
        return queryset


class InvoiceAdmin(ModelAdminBase):
    """Invoice admin. Override following in derived classes:
    - InvoiceSettlementInline with formset derived from AccountEntryInlineFormSet, override clean and call clean_entries()
    - InvoiceItemsInline with formset derived from AccountEntryInlineFormSet, override clean and call clean_entries()
    - inlines = [] set with above mentioned derived classes
    """

    date_hierarchy = "created"
    actions = [
        summarize_invoice_statistics,
        refresh_cached_fields,
        # resend_invoices,
    ]
    # override in derived class
    inlines = [
        InvoiceItemInline,  # TODO: override in app
        InvoiceSettlementInline,  # TODO: override in app
    ]
    list_display = [
        "number",
        "created_brief",
        "sent_brief",
        "due_date_brief",
        "close_date_brief",
        "late_days",
        "amount",
        "paid_amount",
        "unpaid_amount",
    ]
    fields = [
        "type",
        "number",
        "due_date",
        "notes",
        "filename",
        "amount",
        "paid_amount",
        "unpaid_amount",
        "state",
        "overpaid_amount",
        "close_date",
        "late_days",
        "created",
        "last_modified",
        "sent",
    ]
    readonly_fields = [
        "created",
        "last_modified",
        "sent",
        "close_date",
        "created_brief",
        "sent_brief",
        "due_date_brief",
        "close_date_brief",
        "filename",
        "amount",
        "paid_amount",
        "unpaid_amount",
        "state",
        "overpaid_amount",
        "late_days",
    ]
    raw_id_fields = []
    search_fields = [
        "=amount",
        "=filename",
        "=number",
    ]
    list_filter = [
        InvoiceStateFilter,
        InvoiceLateDaysFilter,
    ]
    allow_add = True
    allow_delete = True
    ordering = ("-id",)

    def construct_change_message(self, request, form, formsets, add=False):
        instance = form.instance
        assert isinstance(instance, Invoice)
        instance.update_cached_fields()
        return super().construct_change_message(request, form, formsets, add)

    def _format_date(self, obj) -> str:
        """Short date format.

        Args:
            obj: date or datetime or None

        Returns:
            str
        """
        if obj is None:
            return ""
        if isinstance(obj, datetime):
            obj = obj.date()
        return date_format(obj, "SHORT_DATE_FORMAT")

    def created_brief(self, obj):
        assert isinstance(obj, Invoice)
        return self._format_date(obj.created)

    created_brief.admin_order_field = "created"  # type: ignore
    created_brief.short_description = _("created")  # type: ignore

    def sent_brief(self, obj):
        assert isinstance(obj, Invoice)
        return self._format_date(obj.sent)

    sent_brief.admin_order_field = "sent"  # type: ignore
    sent_brief.short_description = _("sent")  # type: ignore

    def due_date_brief(self, obj):
        assert isinstance(obj, Invoice)
        return self._format_date(obj.due_date)

    due_date_brief.admin_order_field = "due_date"  # type: ignore
    due_date_brief.short_description = _("due date")  # type: ignore

    def close_date_brief(self, obj):
        assert isinstance(obj, Invoice)
        return self._format_date(obj.close_date)

    close_date_brief.admin_order_field = "close_date"  # type: ignore
    close_date_brief.short_description = _("close date")  # type: ignore


def set_as_asset(modeladmin, request, qs):  # pylint: disable=unused-argument
    qs.update(is_asset=True)


def set_as_liability(modeladmin, request, qs):  # pylint: disable=unused-argument
    qs.update(is_asset=False)


class AccountTypeAdmin(ModelAdminBase):
    list_display = [
        "code",
        "name",
        "is_asset",
        "is_liability",
    ]
    actions = [
        set_as_asset,
        set_as_liability,
    ]
    ordering = ("name",)
    allow_add = True
    allow_delete = True

    def is_liability(self, obj):
        return obj.is_liability

    is_liability.short_description = _("is liability")  # type: ignore
    is_liability.boolean = True  # type: ignore


class ContractAdmin(ModelAdminBase):
    list_display = [
        "id",
        "name",
    ]
    ordering = [
        "-id",
    ]
    allow_add = True
    allow_delete = True


def toggle_settlement(modeladmin, request: HttpRequest, queryset: QuerySet):  # pylint: disable=unused-argument
    for e in queryset:
        assert isinstance(e, EntryType)
        e.is_settlement = not e.is_settlement
        e.save()
        admin_log([e], "Toggled settlement flag {}".format("on" if e.is_settlement else "off"), who=request.user)  # type: ignore


def toggle_payment(modeladmin, request: HttpRequest, queryset: QuerySet):  # pylint: disable=unused-argument
    for e in queryset:
        assert isinstance(e, EntryType)
        e.is_payment = not e.is_payment
        e.save()
        admin_log([e], "Toggled payment flag {}".format("on" if e.is_settlement else "off"), who=request.user)  # type: ignore


class EntryTypeAdmin(ModelAdminBase):
    list_display = [
        "id",
        "identifier",
        "name",
        "is_settlement",
        "is_payment",
        "payback_priority",
    ]
    list_filter = (
        "is_settlement",
        "is_payment",
    )
    search_fields = (
        "=code",
        "name",
    )
    actions = [
        toggle_settlement,
        toggle_payment,
    ]
    exclude = ()
    ordering = [
        "name",
    ]
    allow_add = True
    allow_delete = True


class AccountEntrySourceFileAdmin(ModelAdminBase):
    list_display = [
        "id",
        "created",
        "entries_link",
    ]
    date_hierarchy = "created"
    ordering = [
        "-id",
    ]
    fields = [
        "id",
        "name",
        "created",
        "last_modified",
    ]
    search_fields = [
        "=name",
    ]
    readonly_fields = [
        "id",
        "created",
        "name",
        "last_modified",
        "entries_link",
    ]
    allow_add = True
    allow_delete = True

    def entries_link(self, obj):
        if obj and obj.id:
            assert isinstance(obj, AccountEntrySourceFile)
            admin_url = reverse("admin:jacc_accountentry_sourcefile_changelist", args=(obj.id,))
            return format_html("<a href='{}'>{}</a>", mark_safe(admin_url), obj.name)
        return ""

    entries_link.admin_order_field = "name"  # type: ignore
    entries_link.short_description = _("account entry source file")  # type: ignore


class AccountEntryNoteAdmin(ModelAdminBase):
    date_hierarchy = "created"
    fields = [
        "id",
        "account_entry",
        "created",
        "created_by",
        "last_modified",
        "note",
    ]
    raw_id_fields = [
        "account_entry",
        "created_by",
    ]
    readonly_fields = [
        "id",
        "created",
        "created_by",
        "last_modified",
    ]
    write_once_fields = [
        "account_entry",
    ]
    search_fields = [
        "note",
    ]
    list_display = [
        "created",
        "account_entry",
        "note",
        "created_by",
    ]
    list_filter = [
        "created_by",
    ]

    def get_readonly_fields(self, request, obj=None):
        if obj is not None and hasattr(obj, "id") and obj.id:
            return list(self.readonly_fields) + list(self.write_once_fields)
        return self.readonly_fields

    def save_form(self, request, form, change):
        obj = form.instance
        if not change:
            obj.created_by = request.user
        else:
            self.save_account_entry_note(request, obj)
        return form.save(commit=False)

    @staticmethod
    def save_account_entry_note(request, obj: AccountEntryNote):
        if not hasattr(obj, "created_by") or obj.created_by is None:
            obj.created_by = request.user
        else:
            old = AccountEntryNote.objects.all().filter(id=obj.id).first()
            if old is not None:
                assert isinstance(old, AccountEntryNote)
                if old.note != obj.note:
                    obj.created_by = request.user
                    admin_log(
                        [obj, obj.account_entry],
                        "Note id={} modified, previously: {}".format(old.id, old.note),
                        who=request.user,
                    )
        obj.save()

    @staticmethod
    def save_account_entry_note_formset(request, form, formset, change):  # noqa
        assert formset.model == AccountEntryNote
        if formset.model == AccountEntryNote:
            instances = formset.save(commit=False)
            for instance in instances:
                assert isinstance(instance, AccountEntryNote)
                AccountEntryNoteAdmin.save_account_entry_note(request, instance)


add_reverse_charge.short_description = _("Add reverse charge")  # type: ignore
resend_invoices.short_description = _("Re-send invoices")  # type: ignore
refresh_cached_fields.short_description = _("Refresh cached fields")  # type: ignore
summarize_account_entries.short_description = _("Summarize account entries")  # type: ignore
summarize_invoice_statistics.short_description = _("Summarize invoice statistics")  # type: ignore
set_as_asset.short_description = _("set_as_asset")  # type: ignore
set_as_liability.short_description = _("set_as_liability")  # type: ignore

admin.site.register(AccountEntryNote, AccountEntryNoteAdmin)
admin.site.register(Account, AccountAdmin)
admin.site.register(Invoice, InvoiceAdmin)  # TODO: override in app
admin.site.register(AccountEntry, AccountEntryAdmin)  # TODO: override in app
admin.site.register(AccountType, AccountTypeAdmin)
admin.site.register(EntryType, EntryTypeAdmin)
admin.site.register(AccountEntrySourceFile, AccountEntrySourceFileAdmin)



from decimal import Decimal
from typing import Optional, Dict, Any
from django.conf import settings
from django.core.exceptions import ValidationError, ImproperlyConfigured
from django.db import transaction
from django.utils.timezone import now
from django.utils.translation import gettext as _
from jacc.models import AccountEntry, Invoice, EntryType, Account, INVOICE_CREDIT_NOTE, INVOICE_DEFAULT


@transaction.atomic
def settle_invoice(receivables_account: Account, settlement: AccountEntry, invoice: Invoice, cls, **kwargs) -> list:
    """Finds unpaid items in the invoice and generates entries to receivables account.
    Settlement is matched to invoice items based on entry types payback order.
    Generated payment entries have 'parent' field pointing to settlement, so that
    if settlement is (ever) deleted the payment entries will get deleted as well.
    In case of overpayment method generates entry to receivables account without matching invoice settled_item
    (only matching settled_invoice).

    Args:
        receivables_account: Account which receives settled entries of the invoice
        settlement: Settlement to target to unpaid invoice items
        invoice: Invoice to be settled
        cls: Class for generated account entries, e.g. AccountEntry
        **kwargs: Extra attributes for created for generated account entries

    Returns:
        list (generated receivables account entries)
    """
    assert isinstance(invoice, Invoice)
    if not invoice:
        raise ValidationError("Cannot target settlement {} without settled invoice".format(settlement))
    if not receivables_account:
        raise ValidationError("Receivables account missing. Invoice with no rows?")
    if settlement.amount is None:  # nothing to do
        return []
    if settlement.amount < Decimal(0) and invoice.type != INVOICE_CREDIT_NOTE:
        raise ValidationError("Cannot target negative settlement {} to invoice {}".format(settlement, invoice))
    if settlement.amount > Decimal(0) and invoice.type == INVOICE_CREDIT_NOTE:
        raise ValidationError("Cannot target positive settlement {} to credit note {}".format(settlement, invoice))
    if settlement.type is None or not settlement.type.is_settlement:
        raise ValidationError("Cannot settle account entry {} which is not settlement".format(settlement))

    new_payments = []
    remaining = Decimal(settlement.amount)
    timestamp = kwargs.pop("timestamp", settlement.timestamp)
    assert isinstance(invoice, Invoice)
    for item, bal in invoice.get_unpaid_items(receivables_account):
        if invoice.type == INVOICE_DEFAULT:
            if bal > Decimal(0):
                amt = min(remaining, bal)
                ae = cls.objects.create(
                    account=receivables_account,
                    amount=-amt,
                    type=item.type,
                    settled_item=item,
                    settled_invoice=invoice,
                    timestamp=timestamp,
                    description=settlement.description,
                    parent=settlement,
                    **kwargs,
                )
                new_payments.append(ae)
                remaining -= amt
                if remaining <= Decimal(0):
                    break
        elif invoice.type == INVOICE_CREDIT_NOTE:
            if bal < Decimal(0):
                amt = max(remaining, bal)
                ae = cls.objects.create(
                    account=receivables_account,
                    amount=-amt,
                    type=item.type,
                    settled_item=item,
                    settled_invoice=invoice,
                    timestamp=timestamp,
                    description=settlement.description,
                    parent=settlement,
                    **kwargs,
                )
                new_payments.append(ae)
                remaining -= amt
                if remaining >= Decimal(0):
                    break
        else:
            raise NotImplementedError("jacc.settle.settle_assigned_invoice() unimplemented for invoice type {}".format(invoice.type))

    invoice.update_cached_fields()
    return new_payments


@transaction.atomic
def settle_assigned_invoice(receivables_account: Account, settlement: AccountEntry, cls, **kwargs) -> list:
    """Finds unpaid items in the invoice and generates entries to receivables account.
    Settlement is matched to invoice items based on entry types payback order.
    Generated payment entries have 'parent' field pointing to settlement, so that
    if settlement is (ever) deleted the payment entries will get deleted as well.
    In case of overpayment method generates entry to receivables account without matching invoice settled_item
    (only matching settled_invoice).

    Args:
        receivables_account: Account which receives settled entries of the invoice
        settlement: Settlement to target to unpaid invoice items
        cls: Class for generated account entries, e.g. AccountEntry
        **kwargs: Extra attributes for created for generated account entries

    Returns:
        list (generated receivables account entries)
    """
    if settlement.settled_invoice is None:
        raise ValidationError("Cannot target settlement {} without settled invoice".format(settlement))
    return settle_invoice(receivables_account, settlement, settlement.settled_invoice, cls, **kwargs)


@transaction.atomic
def settle_credit_note(  # noqa
    credit_note: Invoice,
    debit_note: Invoice,
    cls,
    account: Account,
    debit_kwargs: Optional[Dict[str, Any]] = None,
    credit_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs
) -> list:
    """Settles credit note. Records settling account entries for both original invoice and the credit note
    (negative entries for the credit note).
    Default timestamp for account entries is current time, can be overriden by kwargs timestamp.

    Args:
        credit_note: Credit note to settle
        debit_note: Invoice to settle
        cls: AccountEntry (derived) class to use for new entries
        account: Settlement account
        debit_kwargs: Optional kwargs for the debit record to cls() instance creation
        credit_kwargs: Optional kwargs for the credit record to cls() instance creation
        **kwargs: Shared variable arguments to cls() instance creation for both records

    Returns:
        list of new settlements
    """
    assert isinstance(credit_note, Invoice)

    if credit_note.type != INVOICE_CREDIT_NOTE:
        raise ValidationError(_("Credit note type incorrect"))
    if debit_note is None:
        raise ValidationError(_("Debit note missing"))
    if debit_note.type != INVOICE_DEFAULT:
        raise ValidationError(_("Debit note type incorrect"))

    credit = -credit_note.get_unpaid_amount()
    balance = debit_note.get_unpaid_amount()

    amt = min(balance, credit)
    amount = kwargs.pop("amount", None)
    if amount is not None:
        if amount > amt:
            raise ValidationError(_("Cannot settle credit note amount which is larger than remaining unpaid balance"))
        amt = amount

    entry_type = kwargs.pop("entry_type", None)
    if entry_type is None:
        if not hasattr(settings, "E_CREDIT_NOTE_RECONCILIATION"):
            err_msg = "settle_credit_note() requires settings.E_CREDIT_NOTE_RECONCILIATION (account entry type code) or entry_type to be pass in kwargs"
            raise ImproperlyConfigured(err_msg)
        entry_type = EntryType.objects.get(code=settings.E_CREDIT_NOTE_RECONCILIATION)
    description = kwargs.pop("description", _("credit.note.reconciliation"))

    pmts = []
    if amt > Decimal(0):
        timestamp = kwargs.pop("timestamp", None)
        if timestamp is None:
            timestamp = now()

        # common args for debit and credit entries
        common_params = {
            "account": account,
            "type": entry_type,
            "timestamp": timestamp,
        }
        for k, v in kwargs.items():
            common_params[k] = v

        # record entry to debit note settlement account
        debit_params = {**common_params}
        if debit_kwargs is not None:
            for k, v in debit_kwargs.items():
                debit_params[k] = v
        pmt1 = cls(
            amount=amt,
            settled_invoice=debit_note,
            description=description + " #{}".format(credit_note.number),
            **debit_params,
        )
        pmt1.clean()
        pmt1.save()
        pmts.append(pmt1)

        # record entry to credit note settlement account
        credit_params = {**common_params}
        if credit_kwargs is not None:
            for k, v in credit_kwargs.items():
                credit_params[k] = v
        pmt2 = cls(
            parent=pmt1,
            amount=-amt,
            settled_invoice=credit_note,
            description=description + " #{}".format(debit_note.number),
            **credit_params,
        )
        pmt2.clean()
        pmt2.save()
        pmts.append(pmt2)

    return pmts



# Generated by Django 3.2 on 2021-07-09 18:10

import django.core.validators
from django.db import migrations
import jutil.modelfields
import re


class Migration(migrations.Migration):

    dependencies = [
        ("jacc", "0024_account_notes"),
    ]

    operations = [
        migrations.AlterField(
            model_name="invoice",
            name="number",
            field=jutil.modelfields.SafeCharField(
                blank=True,
                db_index=True,
                default="",
                error_messages={"invalid": "Only accept letters, numbers, underscores or hyphens."},
                max_length=32,
                validators=[
                    django.core.validators.RegexValidator(
                        re.compile("^[-a-zA-Z0-9_]+\\Z"),
                        "Enter a valid “slug” consisting of letters, numbers, underscores or hyphens.",
                        "invalid",
                    )
                ],
                verbose_name="invoice number",
            ),
        ),
    ]



# Generated by Django 3.2.6 on 2022-02-20 02:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("jacc", "0026_alter_invoice_number"),
    ]

    operations = [
        migrations.AlterField(
            model_name="invoice",
            name="paid_amount",
            field=models.DecimalField(blank=True, db_index=True, decimal_places=2, default=None, max_digits=10, null=True, verbose_name="paid amount"),
        ),
    ]



# Generated by Django 2.2.8 on 2019-12-09 22:31

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("jacc", "0020_auto_20191202_2332"),
    ]

    operations = [
        migrations.AlterField(
            model_name="accountentry",
            name="settled_invoice",
            field=models.ForeignKey(
                blank=True,
                default=None,
                help_text="entry.settled.invoice.help.text",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="jacc.Invoice",
                verbose_name="settled invoice",
            ),
        ),
        migrations.AlterField(
            model_name="accountentry",
            name="settled_item",
            field=models.ForeignKey(
                blank=True,
                default=None,
                help_text="entry.settled.item.help.text",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="settlement_set",
                to="jacc.AccountEntry",
                verbose_name="settled item",
            ),
        ),
        migrations.AlterField(
            model_name="accountentry",
            name="source_file",
            field=models.ForeignKey(
                blank=True,
                default=None,
                help_text="entry.source.file.help.text",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="jacc.AccountEntrySourceFile",
                verbose_name="account entry source file",
            ),
        ),
        migrations.AlterField(
            model_name="accountentry",
            name="source_invoice",
            field=models.ForeignKey(
                blank=True,
                default=None,
                help_text="entry.source.invoice.help.text",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="jacc.Invoice",
                verbose_name="source invoice",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-10-30 19:35
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("jacc", "0003_auto_20171030_1933"),
    ]

    operations = [
        migrations.RenameModel(old_name="AccountEntrySource", new_name="AccountEntrySourceFile"),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-11-02 03:17
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("jacc", "0005_auto_20171030_1958"),
    ]

    operations = [
        migrations.AddField(
            model_name="account",
            name="name",
            field=models.CharField(blank=True, db_index=True, default="", max_length=32, verbose_name="name"),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-10-20 13:31
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("jacc", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="accountentry",
            name="amount",
            field=models.DecimalField(blank=True, decimal_places=2, default=None, max_digits=10, null=True, verbose_name="amount"),
        ),
    ]



from decimal import Decimal
from django.db.models import QuerySet, Sum


def sum_queryset(qs: QuerySet, key: str = "amount", default: Decimal = Decimal(0)) -> Decimal:
    """Returns aggregate sum of queryset 'amount' field.

    Args:
        qs: QuerySet
        key: Field to sum (default: 'amount')
        default: Default value if no results

    Returns:
        Sum of 'amount' field values (coalesced 0 if None)
    """
    res = qs.aggregate(b=Sum(key))["b"]
    return default if res is None else res



"""Double entry accounting system:

A debit is an accounting entry that either increases an asset or expense account,
or decreases a liability or equity account. It is positioned to the left in an accounting entry.
Debit means "left", dividends/expenses/assets/losses increased with debit.

A credit is an accounting entry that either increases a liability or equity account,
or decreases an asset or expense account.
Credit means "right", gains/income/revenues/liabilities/equity increased with credit.
"""

from datetime import datetime, timedelta
from decimal import Decimal
from typing import Optional, Type
from math import floor
from django.core.exceptions import ValidationError
from jacc.helpers import sum_queryset
from django.conf import settings
from django.db import models, transaction
from django.db.models import QuerySet, Q
from django.utils.timezone import now
from jutil.cache import CachedFieldsMixin
from django.utils.translation import gettext_lazy as _
from jutil.format import choices_label
from jutil.modelfields import SafeCharField, SafeTextField

CATEGORY_ANY = ""
CATEGORY_DEBIT = "D"  # "left", dividends/expenses/assets/losses increased with debit
CATEGORY_CREDIT = "C"  # "right", gains/income/revenues/liabilities/equity increased with credit

CATEGORY_TYPE = (
    (CATEGORY_ANY, ""),
    (CATEGORY_DEBIT, _("Debit")),
    (CATEGORY_CREDIT, _("Credit")),
)

INVOICE_NOT_DUE_YET = "N"
INVOICE_DUE = "D"
INVOICE_LATE = "L"
INVOICE_PAID = "P"

INVOICE_STATE = (
    (INVOICE_NOT_DUE_YET, _("Not due yet")),
    (INVOICE_DUE, _("Due")),
    (INVOICE_LATE, _("Late")),
    (INVOICE_PAID, _("Paid")),
)

INVOICE_DEFAULT = "I1"
INVOICE_CREDIT_NOTE = "I2"

INVOICE_TYPE = (
    (INVOICE_DEFAULT, _("Invoice")),
    (INVOICE_CREDIT_NOTE, _("Credit Note")),
)


class AccountEntrySourceFile(models.Model):
    """Account entry source is set for entries based on some event like payment file import"""

    name = SafeCharField(verbose_name=_("name"), max_length=255, db_index=True, blank=True, default="")
    created = models.DateTimeField(verbose_name=_("created"), default=now, db_index=True, editable=False, blank=True)
    last_modified = models.DateTimeField(verbose_name=_("last modified"), auto_now=True, db_index=True, editable=False, blank=True)

    class Meta:
        verbose_name = _("account entry source file")
        verbose_name_plural = _("account entry source files")

    def __str__(self):
        return "[{}] {}".format(self.id, self.name)


class EntryType(models.Model):
    code = SafeCharField(verbose_name=_("code"), max_length=64, db_index=True, unique=True)
    identifier = SafeCharField(verbose_name=_("identifier"), max_length=40, db_index=True, blank=True, default="")
    name = SafeCharField(verbose_name=_("name"), max_length=128, db_index=True, blank=True, default="")
    created = models.DateTimeField(verbose_name=_("created"), default=now, db_index=True, editable=False, blank=True)
    last_modified = models.DateTimeField(verbose_name=_("last modified"), auto_now=True, db_index=True, editable=False, blank=True)
    payback_priority = models.SmallIntegerField(verbose_name=_("payback priority"), default=0, blank=True, db_index=True)
    is_settlement = models.BooleanField(verbose_name=_("is settlement"), default=False, blank=True, db_index=True)
    is_payment = models.BooleanField(verbose_name=_("is payment"), default=False, blank=True, db_index=True)

    class Meta:
        verbose_name = _("entry type")
        verbose_name_plural = _("entry types")

    def __str__(self):
        return "{} ({})".format(self.name, self.code)


class AccountEntryNote(models.Model):
    account_entry = models.ForeignKey("AccountEntry", verbose_name=_("account entry"), related_name="note_set", on_delete=models.CASCADE)
    created = models.DateTimeField(verbose_name=_("created"), default=now, db_index=True, editable=False, blank=True)
    created_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_("created by"),
        editable=False,
        blank=True,
        related_name="accountentrynote_set",
        on_delete=models.CASCADE,
    )
    last_modified = models.DateTimeField(verbose_name=_("last modified"), auto_now=True, editable=False, blank=True)
    note = models.TextField(_("note"))

    class Meta:
        verbose_name = _("account entry note")
        verbose_name_plural = _("account entry notes")


class AccountEntryManager(models.Manager):
    pass


class AccountEntry(models.Model):
    """Single mutation in account state."""

    objects: models.Manager = AccountEntryManager()
    account = models.ForeignKey(
        "Account",
        verbose_name=_("record account"),
        related_name="accountentry_set",
        db_index=True,
        on_delete=models.PROTECT,
    )
    created = models.DateTimeField(verbose_name=_("created"), default=now, db_index=True, editable=False, blank=True)
    last_modified = models.DateTimeField(verbose_name=_("last modified"), auto_now=True, editable=False, blank=True)
    timestamp = models.DateTimeField(verbose_name=_("timestamp"), default=now, db_index=True, blank=True)
    type = models.ForeignKey(
        EntryType,
        verbose_name=_("type"),
        related_name="+",
        on_delete=models.PROTECT,
        null=True,
        default=None,
        blank=True,
    )
    description = SafeCharField(verbose_name=_("description"), max_length=256, default="", blank=True)
    amount = models.DecimalField(verbose_name=_("amount"), max_digits=10, decimal_places=2, blank=True, default=None, null=True, db_index=True)
    source_file = models.ForeignKey(
        AccountEntrySourceFile,
        verbose_name=_("account entry source file"),
        related_name="+",
        null=True,
        default=None,
        blank=True,
        on_delete=models.CASCADE,
        help_text=_("entry.source.file.help.text"),
    )
    source_invoice = models.ForeignKey(
        "Invoice",
        verbose_name=_("source invoice"),
        null=True,
        related_name="+",
        default=None,
        blank=True,
        on_delete=models.CASCADE,
        help_text=_("entry.source.invoice.help.text"),
    )
    settled_invoice = models.ForeignKey(
        "Invoice",
        verbose_name=_("settled invoice"),
        null=True,
        related_name="+",
        default=None,
        blank=True,
        on_delete=models.PROTECT,
        help_text=_("entry.settled.invoice.help.text"),
    )
    settled_item = models.ForeignKey(
        "AccountEntry",
        verbose_name=_("settled item"),
        null=True,
        related_name="settlement_set",
        default=None,
        blank=True,
        on_delete=models.PROTECT,
        help_text=_("entry.settled.item.help.text"),
    )
    parent = models.ForeignKey(
        "AccountEntry",
        verbose_name=_("account.entry.parent"),
        related_name="child_set",
        db_index=True,
        on_delete=models.CASCADE,
        null=True,
        default=None,
        blank=True,
    )
    archived = models.BooleanField(_("archived"), default=False, blank=True)

    class Meta:
        verbose_name = _("account entry")
        verbose_name_plural = _("account entries")
        indexes = [
            models.Index(fields=["account", "created"]),
        ]

    def __str__(self):
        return "[{}] {} {} {}".format(
            self.id,
            self.timestamp.date().isoformat() if self.timestamp else "",
            self.type if self.type else "",
            self.amount,
        )

    def clean(self):
        if self.source_invoice and self.settled_invoice:
            raise ValidationError(
                "Both source_invoice ({}) and settled_invoice ({}) cannot be set same time for account entry ({})".format(
                    self.source_invoice, self.settled_invoice, self
                )
            )

    @property
    def is_parent(self) -> bool:
        """True if this is a parent of some other account entry.

        Returns:
            bool
        """
        return AccountEntry.objects.filter(parent=self).exists()

    @property
    def balance(self) -> Decimal:
        """Returns account balance after this entry.

        Returns:
            Decimal
        """
        return sum_queryset(AccountEntry.objects.filter(account=self.account, timestamp__lte=self.timestamp).exclude(timestamp=self.timestamp, id__gt=self.id))

    balance.fget.short_description = _("balance")  # type: ignore  # pytype: disable=attribute-error


class AccountType(models.Model):
    code = SafeCharField(verbose_name=_("code"), max_length=32, db_index=True, unique=True)
    name = SafeCharField(verbose_name=_("name"), max_length=64, db_index=True, unique=True)
    is_asset = models.BooleanField(verbose_name=_("asset"))
    created = models.DateTimeField(verbose_name=_("created"), default=now, db_index=True, editable=False, blank=True)
    last_modified = models.DateTimeField(verbose_name=_("last modified"), auto_now=True, db_index=True, editable=False, blank=True)

    class Meta:
        verbose_name = _("account type")
        verbose_name_plural = _("account types")

    def __str__(self):
        return str(self.name)

    @property
    def is_liability(self) -> bool:
        return not self.is_asset

    is_liability.fget.short_description = _("liability")  # type: ignore  # pytype: disable=attribute-error


class Account(models.Model):
    """Collects together accounting entries and provides summarizing functionality."""

    type = models.ForeignKey(AccountType, verbose_name=_("type"), related_name="+", on_delete=models.PROTECT)
    name = SafeCharField(verbose_name=_("name"), max_length=64, blank=True, default="", db_index=True)
    currency = SafeCharField(verbose_name=_("currency"), max_length=3, default="EUR", blank=True)
    created = models.DateTimeField(verbose_name=_("created"), default=now, db_index=True, editable=False, blank=True)
    last_modified = models.DateTimeField(verbose_name=_("last modified"), auto_now=True, db_index=True, editable=False, blank=True)
    notes = models.TextField(_("notes"), blank=True, default="")

    class Meta:
        verbose_name = _("account")
        verbose_name_plural = _("accounts")

    def __str__(self):
        return "[{}] {}".format(self.id, self.name or self.type.name)

    def is_asset(self) -> bool:
        return self.type.is_asset

    is_asset.boolean = True  # type: ignore
    is_asset.short_description = _("asset")  # type: ignore

    def is_liability(self) -> bool:
        return self.type.is_liability

    is_liability.boolean = True  # type: ignore
    is_liability.short_description = _("liability")  # type: ignore

    @property
    def balance(self) -> Decimal:
        return sum_queryset(self.accountentry_set.all())

    balance.fget.short_description = _("balance")  # type: ignore  # pytype: disable=attribute-error

    def get_balance(self, t: datetime):
        """
        Returns account balance before specified datetime (excluding entries on the datetime).
        :param t: datetime
        :return: Decimal
        """
        return sum_queryset(self.accountentry_set.all().filter(timestamp__lt=t))

    def needs_settling(self, e: AccountEntry) -> bool:
        """Returns True if all of following conditions are True:
        a) entry has valid amount set
        b) entry type is settlement
        c) entry has been recorded to this account
        d) invoice to be settled has been set
        e) entry has not been settled (=child set empty)

        Args:
            e: AccountEntry (settlement)

        Returns:
            bool
        """
        return bool(e.amount is not None and e.type and e.type.is_settlement and e.account.id == self.id and e.settled_invoice and not e.is_parent)


class InvoiceManager(models.Manager):
    @transaction.atomic
    def update_cached_fields(self, **kw):
        for obj in self.filter(**kw):
            obj.update_cached_fields()


def get_default_due_date():
    return now() + timedelta(days=settings.DEFAULT_DUE_DATE_DAYS) if hasattr(settings, "DEFAULT_DUE_DATE_DAYS") else None


class Invoice(models.Model, CachedFieldsMixin):
    """Invoice model. Typically used as base model for actual app-specific invoice model.

    Convention for naming date/time variables:
    1) date fields are suffixed with _date if they are either plain date fields or interpreted as such (due_date)
    2) natural datetime fields are in past tense, e.g. created, sent (instead of create_date, send_date)

    Note: It is useful sometimes to have full datetime with timezone even for plain dates like due_date,
    because this to be processing to be independent of server, client and invoice time zones.
    """

    objects: models.Manager = InvoiceManager()
    type = SafeCharField(verbose_name=_("type"), max_length=2, db_index=True, default=INVOICE_DEFAULT, blank=True, choices=INVOICE_TYPE)
    number = SafeCharField(verbose_name=_("invoice number"), max_length=32, default="", blank=True, db_index=True)
    created = models.DateTimeField(verbose_name=_("created"), default=now, db_index=True, editable=False, blank=True)
    last_modified = models.DateTimeField(verbose_name=_("last modified"), auto_now=True, db_index=True, editable=False, blank=True)
    sent = models.DateTimeField(verbose_name=_("sent"), db_index=True, default=None, blank=True, null=True)
    due_date = models.DateTimeField(verbose_name=_("due date"), db_index=True, default=get_default_due_date)
    notes = SafeTextField(verbose_name=_("notes"), blank=True, default="")
    filename = SafeCharField(verbose_name=_("filename"), max_length=255, blank=True, default="", db_index=True)
    amount = models.DecimalField(
        verbose_name=_("amount"), max_digits=10, decimal_places=2, default=Decimal("0.00"), blank=True, help_text=_("invoice.amount.help_text")
    )
    paid_amount = models.DecimalField(
        verbose_name=_("paid amount"),
        max_digits=10,
        decimal_places=2,
        blank=True,
        null=True,
        default=Decimal("0.00"),
        db_index=True,
        help_text=_("invoice.paid_amount.help_text"),
    )
    unpaid_amount = models.DecimalField(
        verbose_name=_("unpaid amount"),
        max_digits=10,
        decimal_places=2,
        editable=False,
        blank=True,
        null=True,
        default=None,
        db_index=True,
        help_text=_("invoice.unpaid_amount.help_text"),
    )
    overpaid_amount = models.DecimalField(
        verbose_name=_("overpaid amount"),
        max_digits=10,
        decimal_places=2,
        editable=False,
        blank=True,
        null=True,
        default=Decimal("0.00"),
        db_index=True,
        help_text=_("invoice.overpaid_amount.help_text"),
    )
    close_date = models.DateTimeField(verbose_name=_("close date"), default=None, null=True, blank=True, db_index=True)
    late_days = models.SmallIntegerField(verbose_name=_("late days"), default=None, null=True, blank=True, db_index=True)
    state = SafeCharField(verbose_name=_("state"), max_length=1, blank=True, default="", db_index=True, choices=INVOICE_STATE)
    cached_receivables_account: Optional[Account] = None
    cached_fields = [
        "amount",
        "paid_amount",
        "unpaid_amount",
        "overpaid_amount",
        "close_date",
        "late_days",
        "state",
    ]

    class Meta:
        verbose_name = _("invoice")
        verbose_name_plural = _("invoices")

    def __str__(self):
        return "[{}] {} {}".format(self.id, self.due_date.date().isoformat() if self.due_date else "", self.amount)

    @property
    def receivables_account(self) -> Optional[Account]:
        """Returns receivables account. Receivables account is assumed to be the one were invoice rows were recorded.

        Returns:
            Account or None
        """
        if self.cached_receivables_account is None:
            row = AccountEntry.objects.filter(source_invoice=self).order_by("id").first()
            if row is not None:
                assert isinstance(row, AccountEntry)
                self.cached_receivables_account = row.account
        return self.cached_receivables_account

    @property
    def currency(self) -> str:
        recv = self.receivables_account
        return recv.currency if recv else ""

    def get_entries(self, acc: Account, cls: Type[AccountEntry] = AccountEntry) -> QuerySet:
        """Returns entries related to this invoice on specified account.

        Args:
            acc: Account
            cls: AccountEntry class

        Returns:
            QuerySet
        """
        return cls.objects.filter(Q(account=acc) & (Q(source_invoice=self) | Q(settled_invoice=self))) if acc else cls.objects.none()

    def get_balance(self, acc: Account) -> Decimal:
        """Returns balance of this invoice on specified account.

        Args:
            acc: Account

        Returns:

        """
        return sum_queryset(self.get_entries(acc))

    def get_item_balances(self, acc: Account) -> list:
        """Returns balances of items of the invoice.

        Args:
            acc: Account

        Returns:
            list (AccountEntry, Decimal) in item id order
        """
        items = []
        entries = self.get_entries(acc)
        for item in entries.filter(source_invoice=self).order_by("id"):
            assert isinstance(item, AccountEntry)
            settlements = sum_queryset(entries.filter(settled_item=item))
            bal = item.amount + settlements if item.amount is not None else settlements
            items.append((item, bal))
        return items

    def get_unpaid_items(self, acc: Account) -> list:
        """Returns unpaid items of the invoice in payback priority order.

        Args:
            acc: Account

        Returns:
            list (AccountEntry, Decimal) in payback priority order
        """
        unpaid_items = []
        for item, bal in self.get_item_balances(acc):
            assert isinstance(item, AccountEntry)
            priority = item.type.payback_priority if item.type is not None else 0
            if self.type == INVOICE_DEFAULT:
                if bal > Decimal(0):
                    unpaid_items.append((priority, item, bal))
            elif self.type == INVOICE_CREDIT_NOTE:
                if bal < Decimal(0):
                    unpaid_items.append((priority, item, bal))
            else:
                raise NotImplementedError("jacc.models.Invoice.get_unpaid_items() unimplemented for invoice type {}".format(self.type))
        return [i[1:] for i in sorted(unpaid_items, key=lambda x: x[0])]

    def get_amount(self) -> Decimal:
        return sum_queryset(self.items, "amount")

    @property
    def receivables(self) -> QuerySet:
        acc = self.receivables_account
        if acc is None:
            return AccountEntry.objects.none()
        return self.get_entries(acc)

    @property
    def items(self) -> QuerySet:
        return self.receivables.filter(source_invoice=self)

    def get_paid_amount(self) -> Decimal:
        return self.get_amount() - self.get_unpaid_amount()

    def get_unpaid_amount(self) -> Decimal:
        return sum_queryset(self.receivables)

    def get_overpaid_amount(self) -> Decimal:
        amt = sum_queryset(self.receivables)
        if self.type == INVOICE_CREDIT_NOTE:
            return max(Decimal("0.00"), amt)
        return max(Decimal("0.00"), -amt)

    @property
    def is_paid(self) -> bool:
        if self.unpaid_amount is None:
            return False
        return self.unpaid_amount >= Decimal("0.00") if self.type == INVOICE_CREDIT_NOTE else self.unpaid_amount <= Decimal("0.00")

    is_paid.fget.short_description = _("is paid")  # type: ignore  # pytype: disable=attribute-error

    @property
    def is_due(self) -> bool:
        return not self.is_paid and now() >= self.due_date

    is_due.fget.short_description = _("is due")  # type: ignore  # pytype: disable=attribute-error

    def get_close_date(self) -> Optional[datetime]:
        recv = self.receivables.order_by("-timestamp", "-id")
        first = recv.first()
        if first is None:
            return None
        total = sum_queryset(recv)
        if self.type == INVOICE_CREDIT_NOTE:
            if total >= Decimal("0.00"):
                return first.timestamp
        else:
            if total <= Decimal("0.00"):
                return first.timestamp
        return None

    def get_late_days(self, t: Optional[datetime] = None) -> int:
        t = self.close_date or t
        if t is None:
            t = now()
        return int(floor((t - self.due_date).total_seconds() / 86400.0))

    @property
    def is_late(self) -> bool:
        if self.late_days is None:
            return False
        return not self.is_paid and self.late_days >= settings.LATE_LIMIT_DAYS

    def get_state(self) -> str:
        if self.is_paid:
            return INVOICE_PAID
        t = now()
        if t - self.due_date >= timedelta(days=settings.LATE_LIMIT_DAYS):
            return INVOICE_LATE
        if t >= self.due_date:
            return INVOICE_DUE
        return INVOICE_NOT_DUE_YET

    def get_state_name(self) -> str:
        return choices_label(INVOICE_STATE, self.get_state())

    @property
    def state_name(self) -> str:
        return choices_label(INVOICE_STATE, self.state)

    state_name.fget.short_description = _("state")  # type: ignore  # pytype: disable=attribute-error


class Contract(models.Model):
    """Base class for contracts (e.g. rent contracts, loans, etc.)"""

    created = models.DateTimeField(verbose_name=_("created"), default=now, db_index=True, editable=False, blank=True)
    last_modified = models.DateTimeField(verbose_name=_("last modified"), auto_now=True, db_index=True, editable=False, blank=True)
    name = SafeCharField(verbose_name=_("name"), max_length=128, default="", blank=True, db_index=True)

    class Meta:
        verbose_name = _("contract")
        verbose_name_plural = _("contracts")

    def __str__(self):
        return "[{}]".format(self.id)



# Generated by Django 3.2 on 2021-07-12 15:02

from django.db import migrations
import jutil.modelfields


class Migration(migrations.Migration):

    dependencies = [
        ("jacc", "0025_alter_invoice_number"),
    ]

    operations = [
        migrations.AlterField(
            model_name="invoice",
            name="number",
            field=jutil.modelfields.SafeCharField(blank=True, db_index=True, default="", max_length=32, verbose_name="invoice number"),
        ),
    ]



# Generated by Django 3.1 on 2021-03-11 20:54

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("jacc", "0022_auto_20200629_0346"),
    ]

    operations = [
        migrations.CreateModel(
            name="AccountEntryNote",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        db_index=True,
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name="created",
                    ),
                ),
                ("last_modified", models.DateTimeField(auto_now=True, verbose_name="last modified")),
                ("note", models.TextField(verbose_name="note")),
                (
                    "account_entry",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="note_set",
                        to="jacc.accountentry",
                        verbose_name="account entry",
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        blank=True,
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="accountentrynote_set",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="created by",
                    ),
                ),
            ],
            options={
                "verbose_name": "account entry note",
                "verbose_name_plural": "account entry notes",
            },
        ),
    ]



# Generated by Django 2.0.1 on 2018-02-01 07:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("jacc", "0009_accountentry_archived"),
    ]

    operations = [
        migrations.AlterField(
            model_name="accounttype",
            name="is_asset",
            field=models.BooleanField(verbose_name="asset"),
        ),
    ]



# Generated by Django 2.0.2 on 2018-03-29 10:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("jacc", "0012_auto_20180218_0638"),
    ]

    operations = [
        migrations.AlterField(
            model_name="invoice",
            name="number",
            field=models.BigIntegerField(blank=True, db_index=True, default=None, null=True, verbose_name="invoice number"),
        ),
    ]



# Generated by Django 3.0.7 on 2020-06-29 03:46

from django.db import migrations
import jutil.modelfields


class Migration(migrations.Migration):

    dependencies = [
        ("jacc", "0021_auto_20191209_2231"),
    ]

    operations = [
        migrations.AlterField(
            model_name="account",
            name="currency",
            field=jutil.modelfields.SafeCharField(
                blank=True,
                choices=[("EUR", "EUR"), ("USD", "USD")],
                default="EUR",
                max_length=3,
                verbose_name="currency",
            ),
        ),
        migrations.AlterField(
            model_name="account",
            name="name",
            field=jutil.modelfields.SafeCharField(blank=True, db_index=True, default="", max_length=64, verbose_name="name"),
        ),
        migrations.AlterField(
            model_name="accountentry",
            name="description",
            field=jutil.modelfields.SafeCharField(blank=True, default="", max_length=256, verbose_name="description"),
        ),
        migrations.AlterField(
            model_name="accountentrysourcefile",
            name="name",
            field=jutil.modelfields.SafeCharField(blank=True, db_index=True, default="", max_length=255, verbose_name="name"),
        ),
        migrations.AlterField(
            model_name="accounttype",
            name="code",
            field=jutil.modelfields.SafeCharField(db_index=True, max_length=32, unique=True, verbose_name="code"),
        ),
        migrations.AlterField(
            model_name="accounttype",
            name="name",
            field=jutil.modelfields.SafeCharField(db_index=True, max_length=64, unique=True, verbose_name="name"),
        ),
        migrations.AlterField(
            model_name="contract",
            name="name",
            field=jutil.modelfields.SafeCharField(blank=True, db_index=True, default="", max_length=128, verbose_name="name"),
        ),
        migrations.AlterField(
            model_name="entrytype",
            name="code",
            field=jutil.modelfields.SafeCharField(db_index=True, max_length=64, unique=True, verbose_name="code"),
        ),
        migrations.AlterField(
            model_name="entrytype",
            name="identifier",
            field=jutil.modelfields.SafeCharField(blank=True, db_index=True, default="", max_length=40, verbose_name="identifier"),
        ),
        migrations.AlterField(
            model_name="entrytype",
            name="name",
            field=jutil.modelfields.SafeCharField(blank=True, db_index=True, default="", max_length=128, verbose_name="name"),
        ),
        migrations.AlterField(
            model_name="invoice",
            name="filename",
            field=jutil.modelfields.SafeCharField(blank=True, db_index=True, default="", max_length=255, verbose_name="filename"),
        ),
        migrations.AlterField(
            model_name="invoice",
            name="notes",
            field=jutil.modelfields.SafeTextField(blank=True, default="", verbose_name="notes"),
        ),
        migrations.AlterField(
            model_name="invoice",
            name="number",
            field=jutil.modelfields.SafeCharField(blank=True, db_index=True, default="", max_length=32, verbose_name="invoice number"),
        ),
        migrations.AlterField(
            model_name="invoice",
            name="state",
            field=jutil.modelfields.SafeCharField(
                blank=True,
                choices=[("N", "Not due yet"), ("D", "Due"), ("L", "Late"), ("P", "Paid")],
                db_index=True,
                default="",
                max_length=1,
                verbose_name="state",
            ),
        ),
        migrations.AlterField(
            model_name="invoice",
            name="type",
            field=jutil.modelfields.SafeCharField(
                blank=True,
                choices=[("I1", "Invoice"), ("I2", "Credit Note")],
                db_index=True,
                default="I1",
                max_length=2,
                verbose_name="type",
            ),
        ),
    ]



# Generated by Django 2.0.2 on 2018-02-18 06:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("jacc", "0011_auto_20180208_1146"),
    ]

    operations = [
        migrations.AlterField(
            model_name="accountentry",
            name="amount",
            field=models.DecimalField(
                blank=True,
                db_index=True,
                decimal_places=2,
                default=None,
                max_digits=10,
                null=True,
                verbose_name="amount",
            ),
        ),
        migrations.AlterField(
            model_name="accountentry",
            name="last_modified",
            field=models.DateTimeField(auto_now=True, verbose_name="last modified"),
        ),
    ]
